/*****************************************************************************

  STERM -- A simple multi-threaded terminal communications programming
           example.

  Copyright (C) 1995 by Craig Morrison, All Rights Reserved.

  As with the other modules in this package, if you use it, please
  display my copyright notice in a prominent and conspicuous manner.

  What you see here is the combination of information from many sources:

    Jeff Garzik         OS2TERM
    Jim Gilliland       OS2COM15
    Peter Fitzsimmons   OS2PROG echo (I *think* I got it right this
                                      time Peter.)

  While Jeff had the right idea, his terminal program really smacked of
  MeSsy-DOS programming by the never-ending polling of the keyboard and
  comm port. In short, it required too much CPU time. DosEnterCritSec()
  and DosExitCritSec() should _not_ be made available to application
  programmers. While they make contentions easy to handle and at first
  seem to improve performance, they have an overall impact on the
  application itself. Using them defeats the purpose of writing a
  multi-threaded program in the first place, as they halt all threads in
  a process.

  Jim's code was almost perfect, except for one small problem, it was
  all 16-bit.

  What I have done here is take ideas from both sets of code along with
  a lot of advice from Peter Fitzsimmons (read from the OS2PROG echo) and
  produced a truly OS/2 friendly terminal program. Polling is kept to an
  absolute bare minimum.

  Peter should be happy with this example. :-)

  Lots of good stuff here. Both Mutex and Event semaphores are used to
  handle resource contentions in a friendly manner. 4 different threads
  are used; The main thread handles getting the whole mess up and
  running, then it blocks, waiting for the keyboard thread to end. The
  keyboard thread, keyboard_read(), handles all the keystrokes from the
  local console and determines when we exit. The comm read thread,
  async_read(), takes care of getting data from the port into our
  receive buffer. And lastly, the comm port processing thread,
  async_process(), takes care of handling any special cases of data
  coming from the comm port.

  There is one small problem with the way the receive buffer is handled.
  If the code that removes characters from the buffer gets held up too
  long there is a possibility of the head of the buffer over-running the
  tail. A sufficiently large buffer should keep this from occuring.

  In a perfect world we would also get and save the original device
  control block before we change it. Then just before we exit, restore
  the device control block leaving things the way we found them. But hey,
  it ain't a perfect world and it's not anything MODE won't fix. :-}

 *****************************************************************************/
#define INCL_DOS
#define INCL_DOSDEVIOCTL
#define INCL_SUB
#define INCL_NOPMAPI
#include <os2.h>
#include "str.h"
#include "file_io.h"

#define CBSTACK             8192
#define BAUD_RATE           38400
#define COMM_BUFFER_SIZE    4096
#define COMM_RXBUFFER_SIZE  128
#define MAX_DIALERS         5
#define BANNER              "\x1b[2J\r\nSTERM v1.00, Craig Morrison\r\n" \
                            "Alt-X to exit, Alt-Z for list of commands\r\n\n"
#define USAGE               "Usage: STERM <portnumber>\r\n"
#define NO_KEYPROC          "Can't create keyboard processing thread!\r\n"
#define NO_COMMPROC         "Can't create comm port processing thread!\r\n"
#define NO_COMMTID          "Can't create comm port read thread!\r\n"
#define NO_DCB              "Can't set device control block info!\r\n"
#define NO_LINESET          "Can't set line settings!\r\n"
#define NO_COMMPORT         "Can't open com port!\r\n"
#define CMD_STR             "\r\nCommands:\r\n\nAlt-C  Clear Console.\r\n" \
                            "Alt-D  Dial from directory.\r\n" \
                            "Alt-H  Hang up with DTR.\r\n" \
                            "Alt-X  Quit STERM.\r\nAlt-Z  This list.\r\n" \
                            "\r\n"
#define THANK_YOU           "\r\nThank you for using STERM.\r\nSee ya!\r\n"
#define HANG_UP             "\r\nToggling DTR to hangup..."
#define CLICK               "<CLICK!>\r\n"
#define NO_DIALSTRINGS      "Your dialing directory is empty.\r\n"
#define ERR_TOOHIGH         "Input out of range.\r\n"
#define ERR_NODIALER        "That slot is empty.\r\n"
#define DIAL_PROMPT         "\r\nSelect (ENTER alone aborts): "
#define DIR_HEADER          "\r\n\nDialing Directory:\r\n\n"
#define ANSI_RCP            "\x1b[u"

char buffer[COMM_BUFFER_SIZE];
char dialStr[MAX_DIALERS][128];
int head = 0, tail = 0;
HFILE hCom = NULLHANDLE;
TID kbdProcTID = (TID)-1;
TID comProcTID = (TID)-1;
TID comReceiveTID = (TID)-1;
HMTX comSem = NULLHANDLE;
HEV comReadyHEV = NULLHANDLE;

/* send one character to the comm port */
void charOutPort(char ch)
{
    ULONG BytesWritten;

    DosWrite(hCom, &ch, sizeof(ch), &BytesWritten);
}

/* send an entire string to the comm port */
void strOutPort(PSZ s)
{
    ULONG BytesWritten;

    DosWrite(hCom, s, lenstr(s), &BytesWritten);
}

/*****************************************************************************

  Dial a number from our dialing directory.

 *****************************************************************************/
void DialFromDirectory(void)
{
    char s[128];
    PSZ p;
    ULONG i;
    KBDKEYINFO  ki;

    if (dialStr[0][0]==0)
    {
        VioWrtTTY(NO_DIALSTRINGS, lenstr(NO_DIALSTRINGS), 0L);
        return;
    }
    else
    {
        VioWrtTTY(DIR_HEADER, lenstr(DIR_HEADER), 0L);
        for (i=0; i<MAX_DIALERS; i++)
        {
            if (dialStr[i][0]==0)
                break;
            else
            {
                ultoasc(i, s);
                catstr(s, ")  ");
                VioWrtTTY(s, lenstr(s), 0L);
                VioWrtTTY(dialStr[i], lenstr(dialStr[i]), 0L);
                VioWrtTTY("\r\n", 2, 0L);
            }
        }
        VioWrtTTY(DIAL_PROMPT, lenstr(DIAL_PROMPT), 0L);

        s[0] = 0;
        i = 0;
        for (;;)
        {
            KbdCharIn(&ki, IO_WAIT, 0);
            if (!(ki.fbStatus & 0x02))
            {
                switch(ki.chChar)
                {
                    case '0':
                    case '1':
                    case '2':
                    case '3':
                    case '4':
                    case '5':
                    case '6':
                    case '7':
                    case '8':
                    case '9':
                        if (i<2)
                        {
                            s[i] = ki.chChar;
                            VioWrtTTY(&s[i], 1, 0L);
                            i++;
                        }
                        break;

                    case '\b':
                        if (i)
                        {
                            VioWrtTTY("\b \b", 3, 0L);
                            i--;
                        }
                        break;
                }
                s[i] = 0;

                if (ki.chChar=='\r')
                    break;
            }
        }
        VioWrtTTY("\r\n\n", 2, 0L);

        if (lenstr(s))
        {
            i = asctoul(s);
            if (i>=MAX_DIALERS)
                VioWrtTTY(ERR_TOOHIGH, lenstr(ERR_TOOHIGH), 0L);
            else if (dialStr[i][0]==0)
                VioWrtTTY(ERR_NODIALER, lenstr(ERR_NODIALER), 0L);
            else
            {
                strOutPort("atdt");
                cpystrn(s, dialStr[i], 128);
                p = charinstr(s, ' ');
                if (p)
                    *p = 0;
                strOutPort(s);
                strOutPort("\r");
            }
        }
    }
}

/* Raise or lower DTR signal, based on input */
void dtr(int i)
{
    MODEMSTATUS ms;
    UINT data;

    ms.fbModemOn = i ? DTR_ON : 0;
    ms.fbModemOff = i ? 255 : DTR_OFF;
    DosDevIOCtl(hCom, IOCTL_ASYNC, ASYNC_SETMODEMCTRL, &ms,
                sizeof(ms), NULL, &data, sizeof (data), NULL);
}

/* Removes one character from our port buffer, as an aside, it also watches */
/* for an ANSI DSR sequence, when detected a RCP is sent to the port (many  */
/* BBS software packages use a DSR to detect ANSI capabilities)             */
char charInPort(void)
{
    unsigned long cbOut;
    char c = 0;
    static char dsr = 0;

    DosRequestMutexSem(comSem, SEM_INDEFINITE_WAIT);
    if (head != tail)
    {
        c = buffer[tail++];
        if (tail == COMM_BUFFER_SIZE)
            tail = 0;
        if (c==27)
            dsr = c;
        else if (dsr)
        {
            if ((dsr==27) && (c=='['))
                dsr = c;
            else if ((dsr=='[') && (c=='6'))
                dsr = c;
            else if ((dsr=='6') && (c=='n'))
            {
                DosWrite(hCom, ANSI_RCP, lenstr(ANSI_RCP), &cbOut);
                dsr = 0;
            }
            else
                dsr = 0;
        }
    }
    DosReleaseMutexSem(comSem);

    return(c);
}

/*****************************************************************************

  This is probably the most crucial of the threads. Its job is to read
  incoming data from the comm port and stow it away for later retrieval.

  Note the use of a mutex semaphore to keep the async_process thread from
  modifying the buffer while we are messing with it.

  We post to an event semaphore to unblock the async_process thread if it
  is waiting for something to come in from the port.

 *****************************************************************************/
VOID async_read(PVOID p)
{
    ULONG BytesRead, i;
    char ch[COMM_RXBUFFER_SIZE];

    p = p;  /* shutup compiler */

    for (;;)    /* It keeps on going and going and going and going... */
    {
        /* Try to read in COMM_RXBUFFER_SIZE bytes           */
        /* Since we have the port set to wait-for-something  */
        /* read timeout processing, as soon as the device    */
        /* driver gets some data for us, it coughs it up     */
        if (!DosRead(hCom, ch, COMM_RXBUFFER_SIZE, &BytesRead))
        {
            if (BytesRead)
            {
                /* block here if someone else is mucking with the buffer */
                DosRequestMutexSem(comSem, SEM_INDEFINITE_WAIT);

                /* put all the characters that came in, into the buffer */
                for (i=0; i<BytesRead; i++)
                {
                    buffer[head] = ch[i];
                    head++;
                    if (head == COMM_BUFFER_SIZE)
                      head = 0;
                }

                /* release mutex semaphore so others can get at the buffer */
                DosReleaseMutexSem(comSem);

                /* post to the event sem in case they are blocked waiting */
                DosPostEventSem(comReadyHEV);
            }
        }
    }
}

/*****************************************************************************

  This thread actually does something with the data that has come in from the
  comm port.

 *****************************************************************************/
VOID async_process(PVOID p)
{
    char ch;
    char s[2] = " ";
    ULONG evCount;
    BOOL cbChars;

    p = p;  /* shut compiler */

    for (;;)    /* loop and loop and loop and loop... */
    {
        /* We need to see if anything is waiting in the buffer       */
        /* so we first request the mutex semaphore so the async_read */
        /* will block and leave the buffer pointers alone            */

        DosRequestMutexSem(comSem, SEM_INDEFINITE_WAIT);
        cbChars = (head != tail);
        DosReleaseMutexSem(comSem);

        /* if nothing is waiting, block on an event semaphore until we */
        /* get something in from the comm port                         */
        if (!cbChars)
        {
            DosWaitEventSem(comReadyHEV, SEM_INDEFINITE_WAIT);
            DosResetEventSem(comReadyHEV, &evCount);
        }

        /* get a character and process it */
        switch (ch = charInPort())
        {
            case 0:
            case 255: /* ignore null characters */
                break;

            case 12: /* formfeed char = clear screen */
                VioScrollUp(0, 0, 65535, 65535, 65535, s, 0);
                VioSetCurPos(0, 0, 0);
                break;

            default:    /* everything else gets written out */
                s[0] = ch;
                VioWrtTTY(s, 1, 0L);
                break;
        }
    }
}

/*****************************************************************************

  This thread reads and processes all keyboard input.

 *****************************************************************************/
VOID keyboard_read(PVOID p)
{
    KBDKEYINFO  ki;

    p = p;

    for(;;)
    {
        KbdCharIn(&ki, IO_WAIT, 0);

        if (ki.fbStatus & 0x02) /* extended character? */
        {
            switch (ki.chScan)
            {
                case 0x20:  /* alt-d - dial number */
                    DialFromDirectory();
                    break;

                case 0x2E:  /* alt-c - clear screen */
                    VioWrtTTY("\x1b[2J", 4, 0L);
                    break;

                case 0x23:  /* alt-h - drop DTR to hang up */
                    VioWrtTTY(HANG_UP, lenstr(HANG_UP), 0);
                    dtr(0);
                    DosSleep(500);
                    dtr(1);
                    VioWrtTTY(CLICK, lenstr(CLICK), 0);
                    break;

                case 0x2D:  /* alt-x - exit program */
                    VioWrtTTY(THANK_YOU, lenstr(THANK_YOU), 0);
                    DosExit(EXIT_THREAD, 0);
                    break;

                case 0x2C:  /* alt-z - print menu */
                    VioWrtTTY(CMD_STR, lenstr(CMD_STR), 0);
                    break;
            }
        }
        else
            charOutPort((unsigned char) ki.chChar);
    }
}

/*****************************************************************************

  Sets the bit rate for the comm port.

 *****************************************************************************/
void set_baud(unsigned int rate)
{
    USHORT Baud = (USHORT) rate;

    if ((rate <= 57600) && (rate >= 50))
        DosDevIOCtl(hCom, IOCTL_ASYNC, ASYNC_SETBAUDRATE,
                    &Baud, sizeof(USHORT), NULL, NULL, 0L, NULL);
}

/*****************************************************************************

  Tidies things up before we quit.

 *****************************************************************************/
void closeport(void)
{
    if (comReceiveTID!=(TID)-1)
    {
        DosKillThread(comReceiveTID);
        DosWaitThread(&comReceiveTID, DCWW_WAIT);
    }

    if (comProcTID!=(TID)-1)
    {
        DosKillThread(comProcTID);
        DosWaitThread(&comProcTID, DCWW_WAIT);
    }

    if (kbdProcTID!=(TID)-1)
    {
        DosKillThread(kbdProcTID);
        DosWaitThread(&kbdProcTID, DCWW_WAIT);
    }

    if (comReadyHEV)
        DosCloseEventSem(comReadyHEV);

    if (comSem)
        DosCloseMutexSem(comSem);

    if (hCom)
        DosClose(hCom);

    CleanUpFileIO();
}

/*****************************************************************************

  Opens the comm port, sets the parameters for port communications and
  starts up the threads to handle the port and keyboard.

 *****************************************************************************/
void initport(char port_num)
{
    char s[10] = "COMx";
    APIRET rc;
    ULONG action;
    LINECONTROL lctl;
    DCBINFO dcb;

    InitFileIO();
    DosCreateMutexSem(NULL, &comSem, DC_SEM_SHARED, FALSE);
    DosCreateEventSem(NULL, &comReadyHEV, DC_SEM_SHARED, 0L);

    s[3] = port_num;
    if (DosOpen(s, &hCom, &action, 0L, 0, FILE_OPEN,
                OPEN_ACCESS_READWRITE | OPEN_SHARE_DENYREADWRITE |
                OPEN_FLAGS_FAIL_ON_ERROR, 0L))
    {
        VioWrtTTY(NO_COMMPORT, lenstr(NO_COMMPORT), 0);
        closeport();
        DosExit(EXIT_PROCESS, 1);
    }

    /* set line to no parity, 8 databits, 1 stop bit */
    lctl.bParity = 0;
    lctl.bDataBits = 8;
    lctl.bStopBits = 0;
    lctl.fTransBreak = 0;
    if (DosDevIOCtl(hCom, IOCTL_ASYNC, ASYNC_SETLINECTRL,
                    &lctl, sizeof(LINECONTROL), NULL, NULL, 0L, NULL))
    {
        VioWrtTTY(NO_LINESET, lenstr(NO_LINESET), 0);
        closeport();
        DosExit(EXIT_PROCESS, 1);
    }

    /* set device control block info */
    dcb.usWriteTimeout = 100;               /* one second write timeout */
    dcb.usReadTimeout = 100;                /* one second read timeout */
    dcb.fbCtlHndShake = MODE_DTR_CONTROL;
    dcb.fbFlowReplace = MODE_RTS_CONTROL;
    dcb.fbTimeout = MODE_WAIT_READ_TIMEOUT; /* see the CP Online Reference */
    dcb.bErrorReplacementChar = 0x00;
    dcb.bBreakReplacementChar = 0x00;
    dcb.bXONChar = 0x11;
    dcb.bXOFFChar = 0x13;
    if (DosDevIOCtl(hCom, IOCTL_ASYNC, ASYNC_SETDCBINFO, &dcb,
                    sizeof(DCBINFO), 0L, NULL, 0L, NULL))
    {
        VioWrtTTY(NO_DCB, lenstr(NO_DCB), 0);
        closeport();
        DosExit(EXIT_PROCESS, 1);
    }

    /* Initially our buffer is empty */
    head = tail = 0;

    if (DosCreateThread(&comReceiveTID, (PFNTHREAD)async_read, 0, 0, CBSTACK))
    {
        VioWrtTTY(NO_COMMTID, lenstr(NO_COMMTID), 0);
        closeport();
        DosExit(EXIT_PROCESS, 1);
    }
    if (DosCreateThread(&comProcTID, (PFNTHREAD)async_process, 0, 0, CBSTACK))
    {
        VioWrtTTY(NO_COMMPROC, lenstr(NO_COMMPROC), 0);
        closeport();
        DosExit(EXIT_PROCESS, 1);
    }
    if (DosCreateThread(&kbdProcTID, (PFNTHREAD)keyboard_read, 0, 0, CBSTACK))
    {
        VioWrtTTY(NO_KEYPROC, lenstr(NO_KEYPROC), 0);
        closeport();
        DosExit(EXIT_PROCESS, 1);
    }
    dtr (1);
}

/*****************************************************************************

  Program entry point, here's the scoop on what it does:

    1) ANSI support is turned ON.
    2) Check for a command line arg, we need at least 1.
    3) Initialize the port and start up all of our threads.
    4) Set the bit rate for the port.
    5) Read in our dialing directory.
    6) Wait on the keyboard thread until it dies.
    7) Close the port and tidy up.
    8) Exit with return code 0.

 *****************************************************************************/
void main(int argc, char *argv[])
{
    int i, f;
    PSZ p;
    ULONG ulAction;

    VioSetAnsi(ANSI_ON, 0L);

    VioWrtTTY(BANNER, lenstr(BANNER), 0);

    if (argc != 2)
    {
        VioWrtTTY(USAGE, lenstr(USAGE), 0);
        DosExit(EXIT_PROCESS, 1);
    }

    initport(argv[1][0]);
    set_baud(BAUD_RATE);

    setmem(dialStr, 0, MAX_DIALERS*128);

    f = OpenFile("PHONBOOK.TXT", &ulAction, 0, 0,
                 OPEN_ACTION_OPEN_IF_EXISTS,
                 OPEN_ACCESS_READONLY | OPEN_SHARE_DENYNONE,
                 NULL);
    if (f)
    {
        for (i=0; i<MAX_DIALERS; i++)
        {
            if (LineRead(f, dialStr[i]))
                break;
            else
            {
                p = charinstr(dialStr[i], '\n');
                if (p)
                    *p = 0;
            }
        }
        CloseFile(f);
    }

    DosWaitThread(&kbdProcTID, DCWW_WAIT);
    kbdProcTID = (TID)-1;

    closeport();
    DosExit(EXIT_PROCESS, 0);
}
